# coding=utf-8
from unittest import TestCase
import os
import tempfile

from earthdiagnostics.earthdiags import EarthDiags
from earthdiagnostics.utils import TempFile, Utils
from earthdiagnostics import cdftools
from bscearth.utils.log import Log

import mock
from mock import Mock


class TestEarthDiags(TestCase):

    def setUp(self):
        self.earthdiags = EarthDiags()

    def test_clean(self):
        self.earthdiags.config.scratch_dir = tempfile.mkdtemp()
        os.makedirs(os.path.join(self.earthdiags.config.scratch_dir, 'extra_folder'))

        self.assertTrue(os.path.isdir(self.earthdiags.config.scratch_dir))
        self.earthdiags.clean()
        self.assertFalse(os.path.isdir(self.earthdiags.config.scratch_dir))

    def test_clean_link(self):
        temp_folder = tempfile.mkdtemp()
        scratch_dir = tempfile.mktemp()
        os.symlink(temp_folder, scratch_dir)
        os.makedirs(os.path.join(temp_folder, 'extra_folder'))
        self.earthdiags.config.scratch_dir = scratch_dir

        self.assertTrue(os.path.islink(self.earthdiags.config.scratch_dir))
        self.assertTrue(os.path.isdir(temp_folder))
        self.earthdiags.clean()
        self.assertFalse(os.path.islink(self.earthdiags.config.scratch_dir))
        self.assertFalse(os.path.isdir(temp_folder))

    def test_read_config(self):
        self.earthdiags.config = Mock()
        self.earthdiags.config.scratch_dir = 'scratch_path'
        self.earthdiags.config.cdftools_path = 'cdftools_path'
        self.earthdiags.config.data_convention = Mock()
        self.earthdiags.config.data_convention.lat_name = 'lat'
        self.earthdiags.config.data_convention.lon_name = 'lon'

        self.earthdiags.read_config('config/path')

        self.earthdiags.config.parse.assert_called_once_with('config/path')
        self.assertEqual(os.environ['HDF5_USE_FILE_LOCKING'], 'FALSE')
        self.assertEqual(self.earthdiags.config.scratch_dir, TempFile.scratch_folder)
        self.assertEqual(self.earthdiags.config.cdftools_path, cdftools.path)
        self.assertDictEqual(self.earthdiags.dic_variables,
                             {'nav_lat': 'lat',
                              'nav_lev': 'lev',
                              'nav_lon': 'lon',
                              't': 'time',
                              'time_counter': 'time',
                              'x': 'i',
                              'y': 'j',
                              'z': 'lev'})

    def test_read_config_primavera(self):
        self.earthdiags.config = Mock()
        self.earthdiags.config.scratch_dir = 'scratch_path'
        self.earthdiags.config.cdftools_path = 'cdftools_path'
        self.earthdiags.config.data_convention = Mock()
        self.earthdiags.config.data_convention.lat_name = 'latitude'
        self.earthdiags.config.data_convention.lon_name = 'longitude'

        self.earthdiags.read_config('config/path')

        self.earthdiags.config.parse.assert_called_once_with('config/path')
        self.assertEqual(os.environ['HDF5_USE_FILE_LOCKING'], 'FALSE')
        self.assertEqual(self.earthdiags.config.scratch_dir, TempFile.scratch_folder)
        self.assertEqual(self.earthdiags.config.cdftools_path, cdftools.path)
        self.assertDictEqual(self.earthdiags.dic_variables,
                             {'nav_lat': 'latitude',
                              'nav_lev': 'lev',
                              'nav_lon': 'longitude',
                              't': 'time',
                              'time_counter': 'time',
                              'x': 'i',
                              'y': 'j',
                              'z': 'lev'})

    @mock.patch('earthdiagnostics.earthdiags.EarthDiags.open_documentation')
    def test_parse_args_open_doc(self, open_doc):
        EarthDiags.parse_args(['--doc'])
        open_doc.assert_called_once()

    @mock.patch('earthdiagnostics.earthdiags.EarthDiags.read_config')
    @mock.patch('earthdiagnostics.earthdiags.EarthDiags.clean')
    def test_parse_args_clean(self, clean, read_config):
        EarthDiags.parse_args(['--clean'])
        clean.assert_called_once()
        read_config.assert_called_once_with('diags.conf')

    @mock.patch('earthdiagnostics.earthdiags.EarthDiags.read_config')
    @mock.patch('earthdiagnostics.earthdiags.EarthDiags.report')
    def test_parse_args_report(self, report, read_config):
        EarthDiags.parse_args(['--report'])
        report.assert_called_once()
        read_config.assert_called_once_with('diags.conf')

    @mock.patch('earthdiagnostics.earthdiags.EarthDiags.read_config')
    @mock.patch('earthdiagnostics.earthdiags.EarthDiags.run')
    def test_parse_args_run(self, run, read_config):
        EarthDiags.parse_args([])
        run.assert_called_once()
        read_config.assert_called_once_with('diags.conf')

    @mock.patch('earthdiagnostics.earthdiags.EarthDiags.read_config')
    @mock.patch('earthdiagnostics.earthdiags.EarthDiags.run')
    def test_log_levels(self, run, read_config):
        log_file = tempfile.mktemp()
        for level in ('EVERYTHING', 'DEBUG', 'INFO', 'RESULT', 'USER_WARNING',
                      'WARNING', 'ERROR', 'CRITICAL', 'NO_LOG'):
            EarthDiags.parse_args(['-lc', level, '-lf', level, '-log', log_file])
            self.assertEqual(getattr(Log, level), Log.console_handler.level)
            self.assertEqual(getattr(Log, level), Log.file_handler.level)
        if os.path.isfile(log_file):
            os.remove(log_file)

    @mock.patch('earthdiagnostics.earthdiags.EarthDiags.read_config')
    @mock.patch('earthdiagnostics.earthdiags.EarthDiags.run')
    def test_log_cdo_nco(self, run, read_config):
        EarthDiags.parse_args(['-lc', 'DEBUG'])
        run.assert_called_once()
        read_config.assert_called_once_with('diags.conf')
        self.assertTrue(Utils.cdo.debug)
        self.assertTrue(Utils.nco.debug)
