# coding=utf-8

from mock import Mock
from unittest import TestCase
from earthdiagnostics.variable import CMORTable, VariableAlias, Variable, VariableJsonException, VariableManager
from earthdiagnostics.modelingrealm import ModelingRealms
from earthdiagnostics.constants import Basins
from earthdiagnostics.frequency import Frequencies


class TestCMORTable(TestCase):

    def setUp(self):
        self.frequency = Mock()

    def test_str(self):
        self.assertEqual(str(CMORTable('name', 'm', 'Month YEAR')), 'name')

    def test_repr(self):
        self.assertEqual(repr(CMORTable('name', 'm', 'Month YEAR')), 'name (m, Month YEAR)')

    def test_lt(self):
        self.assertLess(CMORTable('a', 'm', 'Month YEAR'), CMORTable('b', 'm', 'Month YEAR'))


class TestVariableAlias(TestCase):

    def test_str(self):
        alias = VariableAlias('alias')
        self.assertEqual(str(alias), 'alias')
        alias.basin = 'basin'
        self.assertEqual(str(alias), 'alias Basin: basin')
        alias.grid = 'grid'
        self.assertEqual(str(alias), 'alias Basin: basin Grid: grid')

    def test_eq(self):
        alias = VariableAlias('alias')
        self.assertEqual(alias, VariableAlias('alias'))
        alias.basin = 'basin'
        self.assertEqual(alias, VariableAlias('alias', 'basin'))
        alias.grid = 'grid'
        self.assertEqual(alias, VariableAlias('alias', 'basin', 'grid'))
        self.assertNotEqual(alias, None)

    def test_ne(self):
        alias = VariableAlias('alias')
        self.assertNotEqual(alias, VariableAlias('alias2'))
        alias.basin = 'basin'
        self.assertNotEqual(alias, VariableAlias('alias', 'basin2'))
        alias.grid = 'grid'
        self.assertNotEqual(alias, VariableAlias('alias', 'basin', 'grid2'))


class TestVariable(TestCase):

    def test_parse_json(self):
        var = Variable()
        json = {'out_name': 'out_name',
                'standard_name': 'standard_name',
                'long_name': 'long_name',
                'modeling_realm': 'ocean',
                'valid_min': 'valid_min',
                'valid_max': 'valid_max',
                'units': 'units',
                }
        var.parse_json(json, 'out_name')

        self.assertEqual(var.short_name, 'out_name')
        self.assertEqual(var.standard_name, 'standard_name')
        self.assertEqual(var.long_name, 'long_name')

        self.assertEqual(var.valid_min, 'valid_min')
        self.assertEqual(var.valid_max, 'valid_max')
        self.assertEqual(var.units, 'units')
        self.assertEqual(var.priority, 1)

        self.assertEqual(var.domain, ModelingRealms.ocean)

    def test_parse_json_no_out_name(self):
        var = Variable()
        json = {'standard_name': 'standard_name',
                'long_name': 'long_name',
                'modeling_realm': 'ocean',
                'valid_min': 'valid_min',
                'valid_max': 'valid_max',
                'units': 'units',
                }
        with self.assertRaises(VariableJsonException):
            var.parse_json(json, 'out_name')

    def test_parse_json_with_priority(self):
        var = Variable()
        json = {'out_name': 'out_name',
                'standard_name': 'standard_name',
                'long_name': 'long_name',
                'modeling_realm': 'ocean',
                'valid_min': 'valid_min',
                'valid_max': 'valid_max',
                'units': 'units',
                'priority': '2',
                }
        var.parse_json(json, 'out_name')

        self.assertEqual(var.short_name, 'out_name')
        self.assertEqual(var.standard_name, 'standard_name')
        self.assertEqual(var.long_name, 'long_name')

        self.assertEqual(var.valid_min, 'valid_min')
        self.assertEqual(var.valid_max, 'valid_max')
        self.assertEqual(var.units, 'units')
        self.assertEqual(var.priority, 2)

        self.assertEqual(var.domain, ModelingRealms.ocean)

    def test_parse_json_with_primavera_priority(self):
        var = Variable()
        json = {'out_name': 'out_name',
                'standard_name': 'standard_name',
                'long_name': 'long_name',
                'modeling_realm': 'ocean',
                'valid_min': 'valid_min',
                'valid_max': 'valid_max',
                'units': 'units',
                'primavera_priority': '2',
                }
        var.parse_json(json, 'out_name')

        self.assertEqual(var.short_name, 'out_name')
        self.assertEqual(var.standard_name, 'standard_name')
        self.assertEqual(var.long_name, 'long_name')

        self.assertEqual(var.valid_min, 'valid_min')
        self.assertEqual(var.valid_max, 'valid_max')
        self.assertEqual(var.units, 'units')
        self.assertEqual(var.priority, 2)

        self.assertEqual(var.domain, ModelingRealms.ocean)

    def test_get_modelling_realm(self):
        var = Variable()
        domain = var.get_modelling_realm(('ocean',))
        self.assertEqual(ModelingRealms.ocean, domain)

        domain = var.get_modelling_realm(('ocean', 'atmos'))
        self.assertEqual(ModelingRealms.ocean, domain)

        domain = var.get_modelling_realm(('ocean', 'ocnBgchem'))
        self.assertEqual(ModelingRealms.ocnBgchem, domain)

        domain = var.get_modelling_realm(('ocean', 'seaIce'))
        self.assertEqual(ModelingRealms.seaIce, domain)

        domain = var.get_modelling_realm(('atmos', 'atmosChem'))
        self.assertEqual(ModelingRealms.atmosChem, domain)

        domain = var.get_modelling_realm(('land', 'landIce'))
        self.assertEqual(ModelingRealms.landIce, domain)

        domain = var.get_modelling_realm(tuple())
        self.assertIsNone(domain)

    def test_parse_csv(self):
        var = Variable()
        var.parse_csv(['not_used', 'out_name', 'standard_name', 'long_name', 'ocean', 'global', 'units',
                       'valid_min', 'valid_max', 'grid', 'Amon: '])
        self.assertEqual(var.short_name, 'out_name')
        self.assertEqual(var.standard_name, 'standard_name')
        self.assertEqual(var.long_name, 'long_name')

        self.assertEqual(var.valid_min, 'valid_min')
        self.assertEqual(var.valid_max, 'valid_max')
        self.assertEqual(var.units, 'units')
        self.assertEqual(var.grid, 'grid')

        self.assertEqual(var.domain, ModelingRealms.ocean)
        self.assertEqual(var.basin, Basins().Global)

    def test_get_table(self):
        var = Variable()
        var.domain = ModelingRealms.atmos
        table = var.get_table(Frequencies.monthly, 'specs')
        self.assertEqual(table.frequency, Frequencies.monthly)
        self.assertEqual(table.name, 'Amon')
        self.assertEqual(table.date, 'December 2013')

    def test_get_table_added(self):
        convention = Mock()
        convention.name = 'specs'
        var = Variable()
        var.domain = ModelingRealms.atmos
        var.add_table(CMORTable('Amon', Frequencies.monthly, 'December 2013'))
        table = var.get_table(Frequencies.monthly, convention)
        self.assertEqual(table.frequency, Frequencies.monthly)
        self.assertEqual(table.name, 'Amon')
        self.assertEqual(table.date, 'December 2013')

    def test_get_table_not_added(self):
        convention = Mock()
        convention.name = 'specs'
        var = Variable()
        var.domain = ModelingRealms.atmos
        var.add_table(CMORTable('Amon', Frequencies.monthly, 'December 2013'))
        table = var.get_table(Frequencies.daily, convention)
        self.assertEqual(table.frequency, Frequencies.daily)
        self.assertEqual(table.name, 'day')
        self.assertEqual(table.date, 'December 2013')

    def test_get_table_not_matching(self):
        convention = Mock()
        convention.name = 'specs'
        var = Variable()
        with self.assertRaises(ValueError):
            var.get_table(Frequencies.daily, convention)


class TestVariableManager(TestCase):

    def setUp(self):
        self.var_manager = VariableManager()

    def tearDown(self):
        self.var_manager.clean()

    def test_load_primavera(self):
        self.var_manager.load_variables('primavera')
        self.assertTrue(self.var_manager.get_all_variables())

    def test_load_cmip6(self):
        self.var_manager.load_variables('cmip6')
        self.assertTrue(self.var_manager.get_all_variables())

    def test_load_specs(self):
        self.var_manager.load_variables('specs')
        self.assertTrue(self.var_manager.get_all_variables())

    def test_load_preface(self):
        self.var_manager.load_variables('preface')
        self.assertTrue(self.var_manager.get_all_variables())

    def test_bad_load(self):
        with self.assertRaises(Exception):
            self.var_manager.load_variables('badconvention')

    def test_get_variable(self):
        var1 = self._get_var_mock('var1', ['var1_alias'])
        self.var_manager.register_variable(var1)
        self.var_manager.create_aliases_dict()

        self.assertIs(self.var_manager.get_variable('var1'), var1)
        self.assertIs(self.var_manager.get_variable('var1_alias'), var1)

        self.assertIsNone(self.var_manager.get_variable('var2'))
        self.assertIsNone(self.var_manager.get_variable('var2', True))

    def test_get_variable_and_alias(self):
        var1 = self._get_var_mock('var1', ['var1_alias'])
        self.var_manager.register_variable(var1)
        self.var_manager.create_aliases_dict()

        self.assertEqual(self.var_manager.get_variable_and_alias('var1'), (VariableAlias('var1'), var1))
        self.assertEqual(self.var_manager.get_variable_and_alias('var1_alias'), (VariableAlias('var1_alias'), var1))

        self.assertEqual(self.var_manager.get_variable_and_alias('var2'), (None, None))
        self.assertEqual(self.var_manager.get_variable_and_alias('var2', True), (None, None))

    def _get_var_mock(self, name, aliases):
        var1 = Mock()
        var1.short_name = name

        def get_alias_mock(alias):
            alias_mock = Mock()
            alias_mock.alias = alias
            alias_mock.grid = None
            alias_mock.basin = None
            return alias_mock

        var1.known_aliases = [get_alias_mock(alias) for alias in [name] + aliases]
        return var1

    def test_get_all_variables(self):
        var1 = self._get_var_mock('var1', ['var1_alias'])
        var2 = self._get_var_mock('var2', ['var2_alias'])

        self.var_manager.register_variable(var1)
        self.var_manager.register_variable(var2)
        self.var_manager.create_aliases_dict()

        self.assertEqual(self.var_manager.get_all_variables(), [var1, var2])
